from abc import ABC, abstractmethod
from typing import Dict, Any, List
from ml_gym.gym.jobs import GymJob
from dashify.logging.dashify_logging import DashifyLogger, ExperimentInfo


class BluePrint(ABC):
    """ Abstract class that provides a blueprint for creating `AbstractGymJob`
    """

    def __init__(self, model_name: str, dataset_name: str,  epochs: List[int], config: Dict[str, Any], dashify_logging_dir: str,
                 grid_search_id: str, run_id: str, external_injection: Dict[str, Any] = None):

        self.config = config
        self.dashify_logging_dir = dashify_logging_dir
        self.grid_search_id = grid_search_id
        self.run_id = run_id
        self.epochs = epochs
        self.model_name = model_name
        self.dataset_name = dataset_name
        self.external_injection = external_injection if external_injection is not None else {}

    @abstractmethod
    def construct(self) -> GymJob:
        raise NotImplementedError

    def get_experiment_info(self) -> ExperimentInfo:
        experiment_info = DashifyLogger.create_new_experiment(log_dir=self.dashify_logging_dir,
                                                              subfolder_id=self.grid_search_id,
                                                              model_name=self.model_name,
                                                              dataset_name=self.dataset_name,
                                                              run_id=self.run_id)
        DashifyLogger.save_config(config=self.config, experiment_info=experiment_info)
        return experiment_info

    @staticmethod
    @abstractmethod
    def construct_components(config: Dict, external_injection: Dict[str, Any]) -> List[Any]:
        return NotImplementedError



from abc import abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, List
import torch
from ml_gym.batching.batch import DatasetBatch, EvaluationBatchResult, InferenceResultBatch
from ml_gym.data_handling.dataset_loader import DatasetLoader
from ml_gym.gym.inference_component import InferenceComponent
from ml_gym.gym.stateful_components import StatefulComponent
from ml_gym.metrics.metrics import Metric
from ml_gym.models.nn.net import NNModel
from ml_gym.loss_functions.loss_functions import Loss, LossWarmupMixin
import tqdm
from ml_gym.util.logger import LogLevel, ConsoleLogger


class AbstractEvaluator(StatefulComponent):
    @abstractmethod
    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        raise NotImplementedError

    @abstractmethod
    def warm_up(self, model: NNModel, device: torch.device):
        raise NotImplementedError


class Evaluator(AbstractEvaluator):
    def __init__(self, eval_component: 'EvalComponentIF'):
        self.eval_component = eval_component

    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        return self.eval_component.evaluate(model, device)

    def warm_up(self, model: NNModel, device: torch.device):
        self.eval_component.warm_up(model, device)


class EvalComponentIF(StatefulComponent):
    @abstractmethod
    def warm_up(self, model: NNModel, device: torch.device):
        raise NotImplementedError

    @abstractmethod
    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        raise NotImplementedError


class EvalComponent(EvalComponentIF):
    """This thing always comes with batteries included, i.e., datasets, loss functions etc. are all already stored in here."""

    def __init__(self, inference_component: InferenceComponent, metrics: List[Metric],
                 loss_funs: Dict[str, Loss], dataset_loaders: Dict[str, DatasetLoader], train_split_name: str,
                 average_batch_loss: bool = True, show_progress: bool = False):
        self.loss_funs = loss_funs
        self.inference_component = inference_component
        self.metrics = metrics
        self.dataset_loaders = dataset_loaders
        self.average_batch_loss = average_batch_loss
        self.train_split_name = train_split_name
        self.show_progress = show_progress
        self.logger = ConsoleLogger("logger_eval_component")

    def warm_up(self, model: NNModel, device: torch.device):
        def init_loss_funs(batch: InferenceResultBatch):
            for _, loss_fun in self.loss_funs.items():
                if isinstance(loss_fun, LossWarmupMixin):
                    loss_fun.warm_up(batch)
                    loss_fun.finish_warmup()

        if any([isinstance(loss_fun, LossWarmupMixin) for _, loss_fun in self.loss_funs.items()]):
            self.logger.log(LogLevel.INFO, "Running warmup...")
            prediction_batches = self.map_batches(fun=self.forward_batch,
                                                  fun_params={"device": device, "model": model},
                                                  loader=self.dataset_loaders[self.train_split_name],
                                                  show_progress=self.show_progress)
            prediction_batch = InferenceResultBatch.combine(prediction_batches)
            init_loss_funs(prediction_batch)
        else:
            self.logger.log(LogLevel.INFO, "Skipping evaluation warmup. No special loss functions to be initialized.")

    def evaluate(self, model: NNModel, device: torch.device) -> List[EvaluationBatchResult]:
        return [self.evaluate_dataset_split(model, device, split_name, loader) for split_name, loader in self.dataset_loaders.items()]

    def evaluate_dataset_split(self, model: NNModel, device: torch.device, split_name: str, dataset_loader: DatasetLoader) -> EvaluationBatchResult:
        prediction_batches = self.map_batches(fun=self.forward_batch,
                                              loader=dataset_loader,
                                              fun_params={"device": device, "model": model},
                                              show_progress=self.show_progress)
        prediction_batch = InferenceResultBatch.combine(prediction_batches)
        loss_scores = self.calculate_loss_scores(prediction_batch)
        metric_scores = self.calculate_metric_scores(prediction_batch)
        evaluation_result = EvaluationBatchResult(losses=loss_scores,
                                                  metrics=metric_scores,
                                                  dataset_name=dataset_loader.dataset_name,
                                                  split_name=split_name)
        return evaluation_result

    @staticmethod
    def map_batches(fun: Callable[[DatasetBatch, Any], Any], loader: DatasetLoader,
                    fun_params: Dict[str, Any] = None, show_progress: bool = False) -> List[InferenceResultBatch]:
        """
        Applies a function to each batch within a DatasetLoader
        """
        fun_params = fun_params if fun_params is not None else dict()
        # TODO: This loads the entire dataset into memory.
        # We should make this a generator instead to prevent memory overflows.
        if show_progress:
            return [fun(batch, **fun_params) for batch in tqdm.tqdm(loader, desc="Batches processed:")]
        else:
            return [fun(batch, **fun_params) for batch in loader]
        return [fun(batch, **fun_params) for batch in loader]

    def _get_metric_fun(self, identifier: str, target_subscription: Enum, prediction_subscription: Enum,
                        metric_fun: Callable, params: Dict[str, Any]) -> Metric:
        return Metric(identifier, target_subscription, prediction_subscription, metric_fun, params)

    def forward_batch(self, batch: DatasetBatch, model: NNModel, device: torch.device) -> InferenceResultBatch:
        batch.to_device(device)
        model.to(device)
        with torch.no_grad():
            predict_result = self.inference_component.predict(model, batch)
        return predict_result

    def calculate_metric_scores(self, inference_batch: InferenceResultBatch) -> Dict[str, List[float]]:
        return {metric.tag: [metric(inference_batch)] for metric in self.metrics}

    def calculate_loss_scores(self, forward_batch: InferenceResultBatch) -> Dict[str, List[float]]:
        return {loss_key: self._get_batch_loss(loss_fun, forward_batch, self.average_batch_loss)
                for loss_key, loss_fun in self.loss_funs.items()}

    def _get_batch_loss(self, loss_fun: Loss, forward_batch: InferenceResultBatch, averaging: bool = True) -> List[float]:
        loss = loss_fun(forward_batch)
        if averaging:
            # TODO have to check if we actually need to do this since most loss funtions already average ...
            loss = [loss.detach().sum().item() / loss.shape[0]]
        else:
            loss = [loss.detach().tolist()]
        return loss



from abc import ABC, abstractmethod
from ml_gym.gym.jobs import AbstractGymJob


class WorkerIF(ABC):

    @abstractmethod
    def work(job: AbstractGymJob):
        pass


class Worker(WorkerIF):

    def __init__(self):
        pass

    def work(job: AbstractGymJob):
        pass



from typing import Callable, Dict, Any
from functools import partial
from ml_gym.metrics.metrics import PredictionMetric
from ml_gym.batching.batch import InferenceResultBatch


class MetricFactory:

    @staticmethod
    def get_sklearn_metric(metric_key: str, metric_fun: Callable,  params: Dict = None) -> Callable[[InferenceResultBatch], Any]:
        if params is None:
            params = {}
        return partial(PredictionMetric, identifier=metric_key, metric_fun=metric_fun, **params)



from typing import Dict, Any, Type
from ml_gym import create_blueprints
from ml_gym.gym.gym import Gym
from ml_gym.gym.jobs import AbstractGymJob
from ml_gym.validation.validator import ValidatorIF
from ml_gym.blueprints.blue_prints import BluePrint


class GridSearchValidator(ValidatorIF):
    def __init__(self, grid_search_id: str):
        self.grid_search_id = grid_search_id

    def run(self, blue_print_type: Type[BluePrint], gym: Gym, gs_config: Dict[str, Any], num_epochs: int, dashify_logging_path: str):
        blueprints = create_blueprints(blue_print_class=blue_print_type,
                                       run_mode=AbstractGymJob.Mode.TRAIN,
                                       gs_config=gs_config,
                                       dashify_logging_path=dashify_logging_path,
                                       num_epochs=num_epochs, 
                                       grid_search_id=self.grid_search_id)
        gym.add_blue_prints(blueprints)
        gym.run(parallel=True)



from typing import Any, Dict, List, Callable
from data_stack.dataset.iterator import DatasetIteratorIF, CombinedDatasetIterator
from ml_gym.data_handling.postprocessors.postprocessor import LabelMapperPostProcessor, FeatureEncoderPostProcessor, OneHotEncodedTargetPostProcessor
from ml_gym.data_handling.iterators import PostProcessedDatasetIterator
from data_stack.dataset.meta import MetaFactory
from data_stack.dataset.factory import InformedDatasetFactory
from data_stack.dataset.iterator import InformedDatasetIteratorIF
from data_stack.dataset.splitter import SplitterFactory


class ModelGymInformedIteratorFactory(InformedDatasetFactory):

    @staticmethod
    def get_mapped_labels_iterator(identifier: str, iterator: DatasetIteratorIF, mappings: Dict) -> InformedDatasetIteratorIF:
        label_mapper_post_processor = LabelMapperPostProcessor(mappings=mappings,
                                                               target_position=iterator.dataset_meta.target_pos,
                                                               tag_position=iterator.dataset_meta.tag_pos)
        meta = MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta, identifier=identifier)
        return InformedDatasetFactory.get_dataset_iterator(PostProcessedDatasetIterator(iterator, label_mapper_post_processor), meta)

    @staticmethod
    def get_filtered_labels_iterator(identifier: str, iterator: InformedDatasetIteratorIF,
                                     filtered_labels: List[Any]) -> InformedDatasetIteratorIF:
        valid_indices = [i for i in range(len(iterator)) if iterator[i][iterator.dataset_meta.target_pos] in filtered_labels]
        meta = MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta, identifier=identifier)
        return InformedDatasetFactory.get_dataset_iterator_view(iterator, meta, valid_indices)

    @staticmethod
    def get_iterator_view(identifier: str, iterator: InformedDatasetIteratorIF, selection_fun: Callable[[DatasetIteratorIF], List[int]],
                          view_tags: Dict[str, Any]) -> InformedDatasetIteratorIF:
        valid_indices = selection_fun(iterator)
        # valid_indices = list(np.argwhere(valid_mask).flatten())
        meta = MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta, identifier=identifier)
        return InformedDatasetFactory.get_dataset_iterator_view(iterator, meta, valid_indices, view_tags)

    @staticmethod
    def get_feature_encoded_iterators(identifier: str, iterators: Dict[str, InformedDatasetIteratorIF],
                                      feature_encoding_configs: Dict[str, List[Any]]) -> Dict[str, DatasetIteratorIF]:
        sample_position = list(iterators.items())[0][1].dataset_meta.sample_pos
        feature_encoder_post_processor = FeatureEncoderPostProcessor(
            sample_position=sample_position, feature_encoding_configs=feature_encoding_configs)
        feature_encoder_post_processor.fit(iterators)
        return {name: InformedDatasetFactory.get_dataset_iterator(PostProcessedDatasetIterator(iterator, feature_encoder_post_processor),
                                                                  MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta,
                                                                                                             identifier=identifier))
                for name, iterator in iterators.items()}

    @staticmethod
    def get_one_hot_encoded_target_iterators(identifier: str, iterators: Dict[str, InformedDatasetIteratorIF],
                                             target_vector_size: int) -> Dict[str, DatasetIteratorIF]:
        target_position = list(iterators.items())[0][1].dataset_meta.target_pos
        postprocessor = OneHotEncodedTargetPostProcessor(target_vector_size=target_vector_size, target_position=target_position)
        return {name: InformedDatasetFactory.get_dataset_iterator(PostProcessedDatasetIterator(iterator, postprocessor), MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta, identifier=identifier))
                for name, iterator in iterators.items()}

    @staticmethod
    def get_combined_iterators(identifier: str, iterators: Dict[str, Dict[str, InformedDatasetIteratorIF]], combine_configs: Dict) -> Dict[str, InformedDatasetIteratorIF]:
        """Combines iterators.

        Args:
            identifier (str):
            iterators (Dict[str, Dict[str, InformedDatasetIteratorIF]]): Dictionary mapping from iterator_name -> split_name -> iterator
            combine_configs (Dict):

        Returns:
            Dict[str, InformedDatasetIteratorIF]:
        """

        def get_iterators_to_be_combined(iterators: Dict[str, Dict[str, InformedDatasetIteratorIF]], split_config: List):
            return [iterators[element["iterators_name"]][split_name] for element in split_config for split_name in element["splits"]]

        combined_iterators = {}
        for split_config in combine_configs:
            iterator_list = get_iterators_to_be_combined(iterators, split_config["old_splits"])
            meta = MetaFactory.get_dataset_meta_from_existing(dataset_meta=iterator_list[0].dataset_meta, identifier=identifier,
                                                              dataset_name="combined_dataset", dataset_tag=None)
            combined_iterators[split_config["new_split"]] = InformedDatasetFactory.get_dataset_iterator(
                CombinedDatasetIterator(iterator_list), meta)
        return combined_iterators

    @staticmethod
    def get_splitted_iterators(identifier: str, iterators: Dict[str, InformedDatasetIteratorIF], seed: int,
                               split_config: Dict) -> Dict[str, InformedDatasetIteratorIF]:
        def _split(iterator: InformedDatasetIteratorIF, seed: int, split_config: Dict) -> Dict[str, InformedDatasetIteratorIF]:
            names = list(split_config.keys())
            ratios = list(split_config.values())
            splitter = SplitterFactory.get_random_splitter(ratios=ratios, seed=seed)
            splitted_iterators = splitter.split(iterator)
            dataset_metas = [MetaFactory.get_dataset_meta_from_existing(
                iterator.dataset_meta, identifier=identifier, dataset_tag=name) for name in names]
            return {name: InformedDatasetFactory.get_dataset_iterator(splitted_iterators[i], dataset_metas[i]) for i, name in enumerate(names)}

        split_list = []
        for name, iterator in iterators.items():
            if name in split_config.keys():
                split_list.append(_split(iterator, seed, split_config[name]))
            else:
                split_list.append({name: iterator})

        splitted_iterator_dict = {}
        for d in split_list:
            splitted_iterator_dict = {**splitted_iterator_dict, **d}
        return splitted_iterator_dict

    @staticmethod
    def get_in_memory_iterator(identifier: str, iterator: InformedDatasetIteratorIF) -> InformedDatasetIteratorIF:
        meta = MetaFactory.get_dataset_meta_from_existing(iterator.dataset_meta, identifier=identifier)
        return InformedDatasetFactory.get_in_memory_dataset_iterator(iterator, meta)



from abc import ABC, abstractmethod
import numpy as np
from typing import Dict
from sklearn.preprocessing import StandardScaler


class Encoder(ABC):
    @abstractmethod
    def fit(self):
        raise NotImplementedError

    @abstractmethod
    def transform(self):
        raise NotImplementedError

    @abstractmethod
    def get_output_size(self) -> int:
        raise NotImplementedError


class CategoricalEncoder(Encoder):
    def __init__(self):
        super().__init__()
        self.all_values: set = None
        self.value_to_ix: Dict[str, int] = None
        self.ix_to_value: Dict[int, str] = None

    def fit(self, values: np.ndarray):
        self.all_values = np.unique(values.flatten())
        self.value_to_ix = {str(value): ix for ix, value in enumerate(self.all_values)}
        self.ix_to_value = {ix: str(value) for ix, value in enumerate(self.all_values)}

    def transform(self, values: np.ndarray) -> np.ndarray:
        def to_one_hot(value):
            encoded = np.zeros(len(self.all_values))
            encoded[self.value_to_ix[str(value)]] = 1
            return encoded

        if self.all_values is None:
            raise Exception("Please call fit() before transform()")

        transformed = np.array([to_one_hot(value) for value in values.flatten()])
        return transformed

    def get_output_size(self) -> int:
        return len(self.all_values)


class ContinuousEncoder(Encoder):
    def __init__(self):
        super().__init__()
        self.sklearn_encoder = None

    def fit(self, values: np.ndarray):
        self.sklearn_encoder = StandardScaler()
        self.sklearn_encoder.fit(np.expand_dims(values, axis=1))

    def transform(self, values: np.ndarray) -> np.ndarray:
        if self.sklearn_encoder is None:
            raise Exception("Please call fit() before transform()")
        return self.sklearn_encoder.transform(np.expand_dims(values, axis=1))

    def get_output_size(self) -> int:
        return 1



from typing import Dict, Any
from ml_gym.validation.nested_cross_validation import NestedCV
from ml_gym.validation.gs_validator import GridSearchValidator
from ml_gym.blueprints.component_factory import ComponentFactory


class ValidatorFactory:

    @staticmethod
    def get_nested_cv(gs_config: Dict[str, Any], cv_config: Dict[str, Any], grid_search_id: str):
        iterator_key = cv_config["NestedCV"]["iterator_key"]
        split_key = cv_config["NestedCV"]["split_key"]
        component_names = [iterator_key]
        component_factory = ComponentFactory()
        components = component_factory.build_components_from_config(gs_config, component_names)
        iterator = components[iterator_key][split_key]
        return NestedCV(dataset_iterator=iterator, grid_search_id=grid_search_id, **cv_config["NestedCV"]["config"])

    @staticmethod
    def get_gs_validator(grid_search_id: str):
        return GridSearchValidator(grid_search_id=grid_search_id)



from abc import ABC
from typing import Dict, Any, List


class StatefulComponent(ABC):
    """ Provides methods to set and get the temporary state of the object necessary to recreate it.
    """

    def set_state(self, state: Dict[str, Any]):
        """ Sets the the state of all attributes having type `StatefulComponent`
        Args:
            state: state as nested dictionary
        """
        for attr in dir(self):
            if attr.startswith('__'):
                continue
            if self._is_stateful_attribute(attr):
                attr_reference = getattr(self, attr)
                attr_reference.set_state(state[str(attr)])
            elif self._is_list_attribute(attr) and str(attr) in state:
                attr_reference = getattr(self, attr)
                self._set_state_list_attribute(attr_reference, state[str(attr)])
            elif self._is_dict_attribute(attr) and str(attr) in state:
                attr_reference = getattr(self, attr)
                self._set_state_dict_attribute(attr_reference, state[attr])

    def get_state(self) -> Dict[str, Any]:
        """ Returns the state of all attributes having type `StatefulComponent`
        Returns: state as nested dictionary
        """
        state = {}
        for attr in dir(self):
            if attr.startswith('__'):
                continue
            if self._is_stateful_attribute(attr):
                state[str(attr)] = getattr(self, attr).get_state()
            elif self._is_list_attribute(attr):
                state_list = self._get_state_list_attribute(attr)
                if state_list:  # we don't want to have empty lists
                    state[str(attr)] = state_list
            elif self._is_dict_attribute(attr):
                state_dict = self._get_state_dict_attribute(attr)
                if state_dict:  # we don't want to have empty lists
                    state[str(attr)] = state_dict
        return state

    def _is_stateful_attribute(self, attr) -> bool:
        """ Checks if attribute is of type `StatefulComponent`

        Args:
            attr: attribute contained in self

        Returns: True if attribute is of type `StatefulComponent`, False otherwise.
        """
        attr_reference = getattr(self, attr)
        return not attr.startswith('__') and isinstance(attr_reference, StatefulComponent)

    def _get_state_list_attribute(self, list_attr) -> List[Dict[str, Any]]:
        attr_list_reference = getattr(self, list_attr)
        return [element.get_state() for element in attr_list_reference if isinstance(element, StatefulComponent)]

    def _set_state_list_attribute(self, attr_list_reference, states: List[Dict[str, Any]]):
        counter = 0
        for element in attr_list_reference:
            if isinstance(element, StatefulComponent):
                element.set_state(states[counter])
                counter += 1

    def _get_state_dict_attribute(self, dict_attr) -> Dict[str, Any]:
        dict_attr_reference = getattr(self, dict_attr)
        return {key: value.get_state() for key, value in dict_attr_reference.items() if
                isinstance(value, StatefulComponent)}

    def _set_state_dict_attribute(self, dict_attr_reference, states: Dict[str, Any]):
        for key, value in dict_attr_reference.items():
            if isinstance(value, StatefulComponent):
                value.set_state(states[key])

    def _is_list_attribute(self, attr) -> bool:
        """Checks if instance attribute is of type iterable, e.g., list.

        Args:
            attr:

        Returns:

        """
        attr_reference = getattr(self, attr)
        return isinstance(attr_reference, list) and not attr.startswith('__')

    def _is_dict_attribute(self, attr) -> bool:
        attr_reference = getattr(self, attr)
        return isinstance(attr_reference, dict) and not attr.startswith('__')



from setuptools import find_packages, setup

with open("../README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='mlgym',
    version='0.0.15',
    author='Max Luebbering, Rajkumar Ramamurthy',
    description="MLgym, a python framework for distributed machine learning model training in research.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "pytest",
        "pytest-cov",
        "torch",
        "h5py",
        "tqdm",
        "pyyaml",
        "datastack",
        "scipy",
        "dashifyML",
        "scikit-learn"
    ],
    python_requires=">=3.7",
    url="https://github.com/le1nux/mlgym"
)
